/// <reference types="node" />
import { DynamoDBClient } from '@aws-sdk/client-dynamodb';
import { QueryCommandInput } from '@aws-sdk/lib-dynamodb';
import { S3Client } from '@aws-sdk/client-s3';
/**
 * A simple repository for DynamoDB
 * @template DocumentType the type of the document
 * example:
 * ```ts
 * type User {
 *  id: string
 *  name: string
 * }
 * class UserRepository extends SimpleDynamoRepository<User> {}
 * const userRepository = new UserRepository(new DynamoDBClient({ region: 'eu-central-1' }), 'users')
 * ```
 */
export declare abstract class SimpleDynamoRepository<DocumentType> {
    private readonly client;
    private readonly tableName;
    constructor(client: DynamoDBClient, tableName: string);
    /**
     * Find a document by id
     * @param id
     * @throws DocumentNotFoundError if the document does not exist
     * @returns the document
     */
    findOne(id: string): Promise<DocumentType>;
    /**
     * Find a document by the global secondary index
     * @param key the name of the secondary index
     * @param value the value of the secondary index
     * @returns all documents with the given secondary index
     */
    findBy(secondaryKey: string, value: string): Promise<DocumentType[]>;
    /**
     * Find the first document by the global secondary index
     * @param key the name of the secondary index
     * @param value the value of the secondary index
     * @returns the first document with the given secondary index
     */
    findOneBy(secondaryKey: string, value: string): Promise<DocumentType>;
    /**
     * Save a document
     * @param doc
     * @returns the saved document
     */
    save(doc: DocumentType): Promise<DocumentType>;
    /**
     * Delete a document by id
     * @param id
     * @throws DocumentNotFoundError if the document does not exist
     * @returns void
     */
    delete(id: string): Promise<void>;
    /**
     * Find all documents in the table
     * @param limit
     * @param lastEvaluatedId the id of the last document in the previous page
     * @returns all documents in the table
     * example:
     * ```ts
     * let lastEvaluatedId: string | undefined = undefined
     * do {
     *   const users: User[] = await userRepository.findAll(10, lastEvaluatedId)
     *   lastEvaluatedId = users[users.length - 1]?.id || undefined
     * } while (lastEvaluatedId)
     * ```
     */
    findAll(limit?: number, lastEvaluatedId?: string): Promise<DocumentType[]>;
    /**
     * Count all documents in the table
     * @returns the number of documents in the table
     */
    count(): Promise<number>;
    /**
     * Find by query
     * @param queryCommandInput
     * @returns all documents that match the query
     */
    findByQuery(queryCommandInput: Omit<QueryCommandInput, 'TableName'>): Promise<DocumentType[]>;
    /**
     * Find one by query (only returns the first document)
     * @param queryCommandInput
     * @returns all documents that match the query
     */
    findOneByQuery(queryCommandInput: Omit<QueryCommandInput, 'TableName'>): Promise<DocumentType>;
}
/**
 * A simple file interface for S3
 * @example
 * ```ts
 * interface MyFile extends SimpleFile {
 *  name: string
 * }
 * ```
 */
interface SimpleFile {
    path: string;
    fileName: string;
    content: Uint8Array | string | Buffer;
    contentType: string;
    createdAt: string;
    updatedAt: string;
}
/**
 * A simple repository for S3 which is used to save and retrieve files and their metadata
 * @param T the type of the file
 * @param bucketName the name of the bucket
 * @param client the S3 client
 * @example
 * ```ts
 * const s3Client = new S3Client({ region: 'eu-central-1' })
 * const s3Repository = new SimpleS3Repository<SimpleFile>(s3Client, 'my-bucket')
 * ```
 */
export declare class SimpleS3Repository<T extends SimpleFile> {
    private readonly client;
    private readonly bucketName;
    private readonly FIFTEEN_MINUTES;
    constructor(client: S3Client, bucketName: string);
    /**
     * Save a file
     * @param file as SimpleFile
     * @returns void
     * @example
     * ```ts
     * const file: SimpleFile = {
     * path: 'foo/bar/',
     * fileName: 'my-file-name.json',
     * content: 'my-file-content',
     * contentType: 'application/json',
     * createdAt: new Date().toISOString(),
     * updatedAt: new Date().toISOString(),
     * }
     * await s3Repository.save(file)
     * ```
     */
    save(file: T): Promise<void>;
    /**
     * Get a file as string
     * @param filePath the path of the file
     * throws DocumentNotFoundError if the file does not exist
     * @returns the file as string
     */
    getOneAsString(filePath: string): Promise<string>;
    /**
     * Get a file as byte array
     * @param filePath the path of the file
     * throws DocumentNotFoundError if the file does not exist
     * @returns the file as byte array
     */
    getOneAsByteArray(filePath: string): Promise<Uint8Array>;
    /**
     * Get a file as readable stream
     * @param filePath the path of the file
     * throws DocumentNotFoundError if the file does not exist
     * @returns the file as a readable stream
     */
    getOneAsReadableStream(filePath: string): Promise<ReadableStream>;
    private getOne;
    /**
     * Get all which match the prefix
     * @param prefix the prefix
     * @returns all files which match the prefix
     */
    list(prefix: string): Promise<{
        filePath: string;
        lastModified: string;
    }[]>;
    /**
     * Get all metadata for a file
     * @param filePath the path of the file
     * throws DocumentNotFoundError if the file does not exist
     * @returns the metadata of the file
     */
    getOneMetadata(filePath: string): Promise<T>;
    /**
     * Delete a file
     * @param filePath the path of the file
     * @returns void
     */
    deleteOne(filepath: string): Promise<void>;
    /**
     * Delete multiple files
     * @param filePaths the paths of the files
     * @returns void
     */
    delete(filePaths: string[]): Promise<void>;
    /**
     * Get a signed download url
     * @param filePath the path of the file
     * @param expiresIn the time in seconds the url is valid
     * @returns the signed download url
     */
    getSignedDownloadUrl(filePath: string, expiresIn?: number): Promise<string>;
    /**
     * Get a signed upload url
     * @param filePath the path of the file
     * @param expiresIn the time in seconds the url is valid
     * @returns the signed upload url
     */
    getSignedUploadUrl(filePath: string, expiresIn?: number): Promise<string>;
}
export {};
