"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.SimpleS3Repository = exports.SimpleDynamoRepository = void 0;
const lib_dynamodb_1 = require("@aws-sdk/lib-dynamodb");
const client_s3_1 = require("@aws-sdk/client-s3");
const s3_request_presigner_1 = require("@aws-sdk/s3-request-presigner");
class DocumentNotFoundError extends Error {
    constructor(ref) {
        const message = ref ? `Document not found: ${ref}.` : 'Document not found.';
        super(message);
        this.name = 'DocumentNotFoundError';
    }
}
/**
 * A simple repository for DynamoDB
 * @template DocumentType the type of the document
 * example:
 * ```ts
 * type User {
 *  id: string
 *  name: string
 * }
 * class UserRepository extends SimpleDynamoRepository<User> {}
 * const userRepository = new UserRepository(new DynamoDBClient({ region: 'eu-central-1' }), 'users')
 * ```
 */
class SimpleDynamoRepository {
    constructor(client, tableName) {
        this.client = lib_dynamodb_1.DynamoDBDocumentClient.from(client, {
            marshallOptions: { removeUndefinedValues: false },
        });
        this.tableName = tableName;
    }
    /**
     * Find a document by id
     * @param id
     * @throws DocumentNotFoundError if the document does not exist
     * @returns the document
     */
    findOne(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const doc = yield this.client.send(new lib_dynamodb_1.GetCommand({ Key: { id }, TableName: this.tableName }));
            if (!doc.Item)
                throw new DocumentNotFoundError(id);
            return doc.Item;
        });
    }
    /**
     * Find a document by the global secondary index
     * @param key the name of the secondary index
     * @param value the value of the secondary index
     * @returns all documents with the given secondary index
     */
    findBy(secondaryKey, value) {
        return __awaiter(this, void 0, void 0, function* () {
            const queryCommandInput = {
                TableName: this.tableName,
                IndexName: secondaryKey,
                KeyConditionExpression: '#key = :value',
                ExpressionAttributeNames: { '#key': secondaryKey },
                ExpressionAttributeValues: { ':value': value },
            };
            const docs = yield this.client.send(new lib_dynamodb_1.QueryCommand(queryCommandInput));
            return docs.Items;
        });
    }
    /**
     * Find the first document by the global secondary index
     * @param key the name of the secondary index
     * @param value the value of the secondary index
     * @returns the first document with the given secondary index
     */
    findOneBy(secondaryKey, value) {
        return __awaiter(this, void 0, void 0, function* () {
            const queryCommandInput = {
                TableName: this.tableName,
                IndexName: secondaryKey,
                KeyConditionExpression: '#key = :value',
                ExpressionAttributeNames: { '#key': secondaryKey },
                ExpressionAttributeValues: { ':value': value },
            };
            const docs = yield this.client.send(new lib_dynamodb_1.QueryCommand(queryCommandInput));
            if (docs && docs.Items && docs.Items.length > 0)
                return docs.Items[0];
            throw new DocumentNotFoundError();
        });
    }
    /**
     * Save a document
     * @param doc
     * @returns the saved document
     */
    save(doc) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.client.send(new lib_dynamodb_1.PutCommand({
                Item: doc,
                TableName: this.tableName,
            }));
            return doc;
        });
    }
    /**
     * Delete a document by id
     * @param id
     * @throws DocumentNotFoundError if the document does not exist
     * @returns void
     */
    delete(id) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                yield this.client.send(new lib_dynamodb_1.DeleteCommand({
                    Key: { id },
                    TableName: this.tableName,
                    ConditionExpression: 'attribute_exists(id)',
                }));
            }
            catch (error) {
                if (error.name === 'ConditionalCheckFailedException')
                    throw new DocumentNotFoundError(id);
                else
                    throw error;
            }
        });
    }
    /**
     * Find all documents in the table
     * @param limit
     * @param lastEvaluatedId the id of the last document in the previous page
     * @returns all documents in the table
     * example:
     * ```ts
     * let lastEvaluatedId: string | undefined = undefined
     * do {
     *   const users: User[] = await userRepository.findAll(10, lastEvaluatedId)
     *   lastEvaluatedId = users[users.length - 1]?.id || undefined
     * } while (lastEvaluatedId)
     * ```
     */
    findAll(limit = 1000, lastEvaluatedId) {
        return __awaiter(this, void 0, void 0, function* () {
            const scanCommandInput = {
                TableName: this.tableName,
                Limit: limit,
                ExclusiveStartKey: { id: lastEvaluatedId },
            };
            lastEvaluatedId == undefined && delete scanCommandInput.ExclusiveStartKey;
            const docs = yield this.client.send(new lib_dynamodb_1.ScanCommand(scanCommandInput));
            return docs.Items;
        });
    }
    /**
     * Count all documents in the table
     * @returns the number of documents in the table
     */
    count() {
        return __awaiter(this, void 0, void 0, function* () {
            const docs = yield this.client.send(new lib_dynamodb_1.ScanCommand({ TableName: this.tableName, Select: 'COUNT' }));
            return docs.Count || 0;
        });
    }
    /**
     * Find by query
     * @param queryCommandInput
     * @returns all documents that match the query
     */
    findByQuery(queryCommandInput) {
        return __awaiter(this, void 0, void 0, function* () {
            const docs = yield this.client.send(new lib_dynamodb_1.QueryCommand(Object.assign(Object.assign({}, queryCommandInput), { TableName: this.tableName })));
            return docs.Items;
        });
    }
    /**
     * Find one by query (only returns the first document)
     * @param queryCommandInput
     * @returns all documents that match the query
     */
    findOneByQuery(queryCommandInput) {
        return __awaiter(this, void 0, void 0, function* () {
            const docs = yield this.client.send(new lib_dynamodb_1.QueryCommand(Object.assign(Object.assign({}, queryCommandInput), { TableName: this.tableName })));
            if (docs && docs.Items && docs.Items.length > 0)
                return docs.Items[0];
            throw new DocumentNotFoundError();
        });
    }
}
exports.SimpleDynamoRepository = SimpleDynamoRepository;
/**
 * A simple repository for S3 which is used to save and retrieve files and their metadata
 * @param T the type of the file
 * @param bucketName the name of the bucket
 * @param client the S3 client
 * @example
 * ```ts
 * const s3Client = new S3Client({ region: 'eu-central-1' })
 * const s3Repository = new SimpleS3Repository<SimpleFile>(s3Client, 'my-bucket')
 * ```
 */
class SimpleS3Repository {
    constructor(client, bucketName) {
        this.FIFTEEN_MINUTES = 15 * 60;
        this.client = client;
        this.bucketName = bucketName;
    }
    /**
     * Save a file
     * @param file as SimpleFile
     * @returns void
     * @example
     * ```ts
     * const file: SimpleFile = {
     * path: 'foo/bar/',
     * fileName: 'my-file-name.json',
     * content: 'my-file-content',
     * contentType: 'application/json',
     * createdAt: new Date().toISOString(),
     * updatedAt: new Date().toISOString(),
     * }
     * await s3Repository.save(file)
     * ```
     */
    save(file) {
        return __awaiter(this, void 0, void 0, function* () {
            const data = {
                Bucket: this.bucketName,
                Key: file.path + file.fileName,
                Body: file.content,
                ContentType: file.contentType,
                Meta: file,
            };
            yield this.client.send(new client_s3_1.PutObjectCommand(data));
        });
    }
    /**
     * Get a file as string
     * @param filePath the path of the file
     * throws DocumentNotFoundError if the file does not exist
     * @returns the file as string
     */
    getOneAsString(filePath) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield (yield this.getOne(filePath)).transformToString();
        });
    }
    /**
     * Get a file as byte array
     * @param filePath the path of the file
     * throws DocumentNotFoundError if the file does not exist
     * @returns the file as byte array
     */
    getOneAsByteArray(filePath) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield (yield this.getOne(filePath)).transformToByteArray();
        });
    }
    /**
     * Get a file as readable stream
     * @param filePath the path of the file
     * throws DocumentNotFoundError if the file does not exist
     * @returns the file as a readable stream
     */
    getOneAsReadableStream(filePath) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield (yield this.getOne(filePath)).transformToWebStream();
        });
    }
    getOne(filePath) {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield this.client.send(new client_s3_1.GetObjectCommand({ Bucket: this.bucketName, Key: filePath }));
            if (!response.Body)
                throw new DocumentNotFoundError(filePath);
            return response.Body;
        });
    }
    /**
     * Get all which match the prefix
     * @param prefix the prefix
     * @returns all files which match the prefix
     */
    list(prefix) {
        var _a;
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.client.send(new client_s3_1.ListObjectsV2Command({ Bucket: this.bucketName, Prefix: prefix }));
            if (!((_a = res.Contents) === null || _a === void 0 ? void 0 : _a.length))
                return [];
            return res.Contents.map((file) => ({
                filePath: file.Key,
                lastModified: file.LastModified.toISOString(),
            }));
        });
    }
    /**
     * Get all metadata for a file
     * @param filePath the path of the file
     * throws DocumentNotFoundError if the file does not exist
     * @returns the metadata of the file
     */
    getOneMetadata(filePath) {
        return __awaiter(this, void 0, void 0, function* () {
            const response = yield this.client.send(new client_s3_1.HeadObjectCommand({ Bucket: this.bucketName, Key: filePath }));
            console.log({ response });
            return response.Metadata;
        });
    }
    /**
     * Delete a file
     * @param filePath the path of the file
     * @returns void
     */
    deleteOne(filepath) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.delete([filepath]);
        });
    }
    /**
     * Delete multiple files
     * @param filePaths the paths of the files
     * @returns void
     */
    delete(filePaths) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!filePaths.length)
                return;
            const Objects = filePaths.map((filePath) => ({ Key: filePath }));
            yield this.client.send(new client_s3_1.DeleteObjectsCommand({ Bucket: this.bucketName, Delete: { Objects } }));
        });
    }
    /**
     * Get a signed download url
     * @param filePath the path of the file
     * @param expiresIn the time in seconds the url is valid
     * @returns the signed download url
     */
    getSignedDownloadUrl(filePath, expiresIn = this.FIFTEEN_MINUTES) {
        return __awaiter(this, void 0, void 0, function* () {
            const command = new client_s3_1.GetObjectCommand({
                Bucket: this.bucketName,
                Key: filePath,
            });
            return yield (0, s3_request_presigner_1.getSignedUrl)(this.client, command, { expiresIn });
        });
    }
    /**
     * Get a signed upload url
     * @param filePath the path of the file
     * @param expiresIn the time in seconds the url is valid
     * @returns the signed upload url
     */
    getSignedUploadUrl(filePath, expiresIn = this.FIFTEEN_MINUTES) {
        return __awaiter(this, void 0, void 0, function* () {
            const command = new client_s3_1.PutObjectCommand({
                Bucket: this.bucketName,
                Key: filePath,
            });
            return yield (0, s3_request_presigner_1.getSignedUrl)(this.client, command, { expiresIn });
        });
    }
}
exports.SimpleS3Repository = SimpleS3Repository;
