import { DynamoDBClient } from '@aws-sdk/client-dynamodb'
import {
  DeleteCommand,
  DynamoDBDocumentClient,
  QueryCommand,
  ScanCommand,
  GetCommand,
  PutCommand,
  QueryCommandInput,
  ScanCommandInput,
} from '@aws-sdk/lib-dynamodb'

import {
  S3Client,
  ListObjectsV2Command,
  GetObjectCommand,
  DeleteObjectsCommand,
  PutObjectCommand,
  _Object,
  HeadObjectCommand,
} from '@aws-sdk/client-s3'
import { getSignedUrl } from '@aws-sdk/s3-request-presigner'

class DocumentNotFoundError extends Error {
  constructor(ref?: string) {
    const message = ref ? `Document not found: ${ref}.` : 'Document not found.'
    super(message)
    this.name = 'DocumentNotFoundError'
  }
}

/**
 * A simple repository for DynamoDB
 * @template DocumentType the type of the document
 * example:
 * ```ts
 * type User {
 *  id: string
 *  name: string
 * }
 * class UserRepository extends SimpleDynamoRepository<User> {}
 * const userRepository = new UserRepository(new DynamoDBClient({ region: 'eu-central-1' }), 'users')
 * ```
 */
export abstract class SimpleDynamoRepository<DocumentType> {
  private readonly client: DynamoDBDocumentClient
  private readonly tableName: string

  constructor(client: DynamoDBClient, tableName: string) {
    this.client = DynamoDBDocumentClient.from(client, {
      marshallOptions: { removeUndefinedValues: false },
    })
    this.tableName = tableName
  }

  /**
   * Find a document by id
   * @param id
   * @throws DocumentNotFoundError if the document does not exist
   * @returns the document
   */
  async findOne(id: string): Promise<DocumentType> {
    const doc = await this.client.send(new GetCommand({ Key: { id }, TableName: this.tableName }))

    if (!doc.Item) throw new DocumentNotFoundError(id)

    return doc.Item as DocumentType
  }

  /**
   * Find a document by the global secondary index
   * @param key the name of the secondary index
   * @param value the value of the secondary index
   * @returns all documents with the given secondary index
   */
  async findBy(secondaryKey: string, value: string): Promise<DocumentType[]> {
    const queryCommandInput: QueryCommandInput = {
      TableName: this.tableName,
      IndexName: secondaryKey,
      KeyConditionExpression: '#key = :value',
      ExpressionAttributeNames: { '#key': secondaryKey },
      ExpressionAttributeValues: { ':value': value },
    }
    const docs = await this.client.send(new QueryCommand(queryCommandInput))

    return docs.Items as DocumentType[]
  }

  /**
   * Find the first document by the global secondary index
   * @param key the name of the secondary index
   * @param value the value of the secondary index
   * @returns the first document with the given secondary index
   */
  async findOneBy(secondaryKey: string, value: string): Promise<DocumentType> {
    const queryCommandInput: QueryCommandInput = {
      TableName: this.tableName,
      IndexName: secondaryKey,
      KeyConditionExpression: '#key = :value',
      ExpressionAttributeNames: { '#key': secondaryKey },
      ExpressionAttributeValues: { ':value': value },
    }
    const docs = await this.client.send(new QueryCommand(queryCommandInput))

    if (docs && docs.Items && docs.Items.length > 0) return docs.Items[0] as DocumentType

    throw new DocumentNotFoundError()
  }

  /**
   * Save a document
   * @param doc
   * @returns the saved document
   */
  async save(doc: DocumentType): Promise<DocumentType> {
    await this.client.send(
      new PutCommand({
        Item: doc as Record<string, any>,
        TableName: this.tableName,
      }),
    )

    return doc
  }

  /**
   * Delete a document by id
   * @param id
   * @throws DocumentNotFoundError if the document does not exist
   * @returns void
   */
  async delete(id: string): Promise<void> {
    try {
      await this.client.send(
        new DeleteCommand({
          Key: { id },
          TableName: this.tableName,
          ConditionExpression: 'attribute_exists(id)',
        }),
      )
    } catch (error: any) {
      if (error.name === 'ConditionalCheckFailedException') throw new DocumentNotFoundError(id)
      else throw error
    }
  }

  /**
   * Find all documents in the table
   * @param limit
   * @param lastEvaluatedId the id of the last document in the previous page
   * @returns all documents in the table
   * example:
   * ```ts
   * let lastEvaluatedId: string | undefined = undefined
   * do {
   *   const users: User[] = await userRepository.findAll(10, lastEvaluatedId)
   *   lastEvaluatedId = users[users.length - 1]?.id || undefined
   * } while (lastEvaluatedId)
   * ```
   */
  async findAll(limit = 1000, lastEvaluatedId?: string): Promise<DocumentType[]> {
    const scanCommandInput: ScanCommandInput = {
      TableName: this.tableName,
      Limit: limit,
      ExclusiveStartKey: { id: lastEvaluatedId },
    }

    lastEvaluatedId == undefined && delete scanCommandInput.ExclusiveStartKey

    const docs = await this.client.send(new ScanCommand(scanCommandInput))

    return docs.Items as DocumentType[]
  }

  /**
   * Count all documents in the table
   * @returns the number of documents in the table
   */
  async count(): Promise<number> {
    const docs = await this.client.send(new ScanCommand({ TableName: this.tableName, Select: 'COUNT' }))

    return docs.Count || 0
  }

  /**
   * Find by query
   * @param queryCommandInput
   * @returns all documents that match the query
   */
  async findByQuery(queryCommandInput: Omit<QueryCommandInput, 'TableName'>): Promise<DocumentType[]> {
    const docs = await this.client.send(
      new QueryCommand({
        ...queryCommandInput,
        TableName: this.tableName,
      }),
    )

    return docs.Items as DocumentType[]
  }

  /**
   * Find one by query (only returns the first document)
   * @param queryCommandInput
   * @returns all documents that match the query
   */
  async findOneByQuery(queryCommandInput: Omit<QueryCommandInput, 'TableName'>): Promise<DocumentType> {
    const docs = await this.client.send(
      new QueryCommand({
        ...queryCommandInput,
        TableName: this.tableName,
      }),
    )

    if (docs && docs.Items && docs.Items.length > 0) return docs.Items[0] as DocumentType

    throw new DocumentNotFoundError()
  }
}

/**
 * A simple file interface for S3
 * @example
 * ```ts
 * interface MyFile extends SimpleFile {
 *  name: string
 * }
 * ```
 */
interface SimpleFile {
  path: string
  fileName: string
  content: Uint8Array | string | Buffer
  contentType: string
  createdAt: string
  updatedAt: string
}

/**
 * A simple repository for S3 which is used to save and retrieve files and their metadata
 * @param T the type of the file
 * @param bucketName the name of the bucket
 * @param client the S3 client
 * @example
 * ```ts
 * const s3Client = new S3Client({ region: 'eu-central-1' })
 * const s3Repository = new SimpleS3Repository<SimpleFile>(s3Client, 'my-bucket')
 * ```
 */
export class SimpleS3Repository<T extends SimpleFile> {
  private readonly client: S3Client
  private readonly bucketName: string
  private readonly FIFTEEN_MINUTES = 15 * 60

  constructor(client: S3Client, bucketName: string) {
    this.client = client
    this.bucketName = bucketName
  }

  /**
   * Save a file
   * @param file as SimpleFile
   * @returns void
   * @example
   * ```ts
   * const file: SimpleFile = {
   * path: 'foo/bar/',
   * fileName: 'my-file-name.json',
   * content: 'my-file-content',
   * contentType: 'application/json',
   * createdAt: new Date().toISOString(),
   * updatedAt: new Date().toISOString(),
   * }
   * await s3Repository.save(file)
   * ```
   */
  async save(file: T): Promise<void> {
    const data = {
      Bucket: this.bucketName,
      Key: file.path + file.fileName,
      Body: file.content,
      ContentType: file.contentType,
      Meta: file,
    }

    await this.client.send(new PutObjectCommand(data))
  }

  /**
   * Get a file as string
   * @param filePath the path of the file
   * throws DocumentNotFoundError if the file does not exist
   * @returns the file as string
   */
  async getOneAsString(filePath: string): Promise<string> {
    return await (await this.getOne(filePath)).transformToString()
  }

  /**
   * Get a file as byte array
   * @param filePath the path of the file
   * throws DocumentNotFoundError if the file does not exist
   * @returns the file as byte array
   */
  async getOneAsByteArray(filePath: string): Promise<Uint8Array> {
    return await (await this.getOne(filePath)).transformToByteArray()
  }

  /**
   * Get a file as readable stream
   * @param filePath the path of the file
   * throws DocumentNotFoundError if the file does not exist
   * @returns the file as a readable stream
   */
  async getOneAsReadableStream(filePath: string): Promise<ReadableStream> {
    return await (await this.getOne(filePath)).transformToWebStream()
  }

  private async getOne(filePath: string) {
    const response = await this.client.send(new GetObjectCommand({ Bucket: this.bucketName, Key: filePath }))

    if (!response.Body) throw new DocumentNotFoundError(filePath)

    return response.Body
  }

  /**
   * Get all which match the prefix
   * @param prefix the prefix
   * @returns all files which match the prefix
   */
  async list(prefix: string): Promise<{ filePath: string; lastModified: string }[]> {
    const res = await this.client.send(new ListObjectsV2Command({ Bucket: this.bucketName, Prefix: prefix }))

    if (!res.Contents?.length) return []

    return res.Contents.map((file) => ({
      filePath: file.Key!,
      lastModified: file.LastModified!.toISOString(),
    }))
  }

  /**
   * Get all metadata for a file
   * @param filePath the path of the file
   * throws DocumentNotFoundError if the file does not exist
   * @returns the metadata of the file
   */
  async getOneMetadata(filePath: string): Promise<T> {
    const response = await this.client.send(new HeadObjectCommand({ Bucket: this.bucketName, Key: filePath }))

    console.log({ response })

    return response.Metadata as unknown as T
  }

  /**
   * Delete a file
   * @param filePath the path of the file
   * @returns void
   */
  async deleteOne(filepath: string): Promise<void> {
    await this.delete([filepath])
  }

  /**
   * Delete multiple files
   * @param filePaths the paths of the files
   * @returns void
   */
  async delete(filePaths: string[]): Promise<void> {
    if (!filePaths.length) return

    const Objects = filePaths.map((filePath) => ({ Key: filePath }))
    await this.client.send(new DeleteObjectsCommand({ Bucket: this.bucketName, Delete: { Objects } }))
  }

  /**
   * Get a signed download url
   * @param filePath the path of the file
   * @param expiresIn the time in seconds the url is valid
   * @returns the signed download url
   */
  async getSignedDownloadUrl(filePath: string, expiresIn = this.FIFTEEN_MINUTES): Promise<string> {
    const command = new GetObjectCommand({
      Bucket: this.bucketName,
      Key: filePath,
    })

    return await getSignedUrl(this.client, command, { expiresIn })
  }

  /**
   * Get a signed upload url
   * @param filePath the path of the file
   * @param expiresIn the time in seconds the url is valid
   * @returns the signed upload url
   */
  async getSignedUploadUrl(filePath: string, expiresIn = this.FIFTEEN_MINUTES): Promise<string> {
    const command = new PutObjectCommand({
      Bucket: this.bucketName,
      Key: filePath,
    })

    return await getSignedUrl(this.client, command, { expiresIn })
  }
}
